//  ComputationalAssemblerBasic.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;

import java.io.*;
import java.util.*;

import java.lang.reflect.Method;

/**
 * The ComputationalAssemblerBasic class instances make 
 * computational processes; we have a computationa assembler instance for
 * each unit
 *
 * THIS CLASS IS NOT USED DIRECTLY, BUT VIA THE INHERITING CLASS
 * ComputationalAssembler.java
 *
 * @author Pietro Terna; Carlo Desotgiu made the steps 1001 and 1002
 */
public class ComputationalAssemblerBasic extends SwarmObjectImpl{

    /** the memory matrixes used in computations; we can add a lot of these */
    MemoryMatrix mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7, mm8, mm9;
    /** the current layer */
    public int layer;

    /** the unit for which we are assembling for */
    public Unit myUnit;
    /** the list of the order requiring computations */
    public ListImpl waitingList;
    /** a computational specification set */
    ComputationalSpecificationSet pendingComputationalSpecificationSet;
    /** the status of a specific computation */
    public boolean done;

    // object to be addressed
    OrderDistiller myOrderDistiller;
    ESFrameModelSwarm eSFrameModelSwarm;

    /**
     * the constructor for ComputationalAssembler
     */
    public ComputationalAssemblerBasic (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
    {
	// Call the constructor for the parent class.
	super(aZone);

	myOrderDistiller = od;
	eSFrameModelSwarm=mo;

	// the list of orders with computational processes to be assembled
	waitingList = new ListImpl (getZone());

    }

    /** setting the unit we are assembling for */
    public void setUnit(Unit u){
	myUnit = u;
    }

    /** adding an order to the waitingList */
    public boolean setComputationalWaitingList(Order anOrder){
	waitingList.addLast(anOrder);
	return true;
    }

    /** making computations and eliminating the orders from our
     *  waitingList */
    public void checkingComputationsAndFreeingOrders(){
	int n, nmax, t;
	Order anOrder;

	nmax = waitingList.getCount();
	if(nmax>0)
	    for (n=0;n<nmax;n++)
		{
		    anOrder = (Order) waitingList.removeFirst();

		    pendingComputationalSpecificationSet = 
			(ComputationalSpecificationSet)
			 anOrder.getPendingComputationalSpecificationSet();

		    if(pendingComputationalSpecificationSet == null){
			System.out.println("Internal error in" +
				       " computationalAssembler of unit " +
				       myUnit.getUnitNumber() +
				       " - order # " + 
				       anOrder.getOrderNumber() +  
				       "waiting for computations "+
				       "has no computational specifications.");
			MyExit.exit(1);
		    }

		    // computations
		    done=false;
		    if(StartESFrame.verbose)
				System.out.println
				    ("Making computations for order # " +
				     pendingComputationalSpecificationSet.
				     getOrderNumber() + " of type " +
				     pendingComputationalSpecificationSet.
				     getComputationType() + " with " + 
				     pendingComputationalSpecificationSet.
				     getNumberOfMemoryMatrixesToBeUsed() +
				     " matrixes");

				t=pendingComputationalSpecificationSet.
				    getComputationType();

				// transforming the code contained in t (a
				// number in the range -1000 -1999) in a 
				// method label (c1000 c1999) via reflection
				// methods
				try{
				    Class c = this.getClass();
				    Method m = c.getMethod("c"+(-1*t),null);
				    m.invoke(this, null);
				}
				catch (Exception e) {
				    System.out.println(e);
				    System.out.println("Invalid computational"+
						       " code in order: " +
						       anOrder.
						       getOrderNumber());
				    MyExit.exit(1);
				}


		    // exiting
		    if(done)
			{
			    if(StartESFrame.verbose)
				System.out.println
				    ("Made computations for order # " +
				     pendingComputationalSpecificationSet.
				     getOrderNumber());
			    pendingComputationalSpecificationSet.setDone();
			    // the last set at present has not effect
			    // it is introduced for possible future uses
			}

		    else waitingList.addLast(anOrder);

		}
    }

    /**
     * return the waiting list length
     */
    public int getWaitingListLength()
    {
	return waitingList.getCount();
    }

    /**
     * checking if an order is in the waitingList
     */
    public boolean thisOrderIsInTheWaitingList(Order o){

	// the result is true if o is a member of waitingList; false in
	// the opposite case
	return waitingList.contains(o);
    }

    /** removing an order form the waintingList */
    public void removeThisOrderFromTheWaitingList(Order o){
	waitingList.remove(o);
    }

     /** computational operations with code -1995 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code place a 1 in position 1,1 of the
     *  unique received matrix and set the status to done
     */
    public void c1995(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1)
	    {
		System.out.println("Code -1995 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	mm0.setValue(layer,1,1,1.0);
	mm0.print();

	done=true;
    } // end c1995

    /** computational operations with code -1996 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code place a 1 in position 0,0 of the
     *  unique received matrix and set the status to done
     */
    public void c1996(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1)
	    {
		System.out.println("Code -1996 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	mm0.setValue(layer,0,0,1.0);
	mm0.print();

	done=true;
    } // end c1996

    /** computational operations with code -1997 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code place a 1 in position 1,0 of the
     *  unique received matrix and set the status to done
     */
    public void c1997(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1)
	    {
		System.out.println("Code -1997 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	mm0.setValue(layer,1,0,1.0);
	mm0.print();

	done=true;
    } // end c1997

    /** computational operations with code -1998 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code place a number in position 0,0 of the
     *  unique received matrix and set the status to done
     */
    public void c1998(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1)
	    {
		System.out.println("Code -1998 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	mm0.setValue(layer,0,0,1.0);
	mm0.print();

	done=true;
    } // end c1998

    /** computational operations with code -1999 (a code for the checking
     *  phase of the program <br><br>
     *
     *  this computational code verifies position 0,0 of the three
     *  received matrixes; only if these positions are all not empty
     *  the code empties them and set the status to done
     */
    public void c1999(){
	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=3)
	    {
		System.out.println("Code -1999 requires three matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	mm2=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(2);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	if(! (mm0.getEmpty(layer,0,0) || mm1.getEmpty(layer,0,0) 
	                              || mm2.getEmpty(layer,0,0) ) )
	    {
		mm0.setEmpty(layer,0,0);
		mm1.setEmpty(layer,0,0);
		mm2.setEmpty(layer,0,0);
		done=true;
	    }

    } // end c1999


    // computational steps 1001 and 1002 to launch recipes from recipes

    /**
     * computational operations with code -1001
     *
     * this computational code places the numbers 
     * of the recipes to launch from other recipes, via the c step 1002,
     * in position 0,0 of the received matrixes and set the status to done.
     *
     * the number of the matrixes used here must be less or equal to totalMemoryMatrixNumber
     * (the first m matrixes of these ones); beside this, m has to be less than 10
     */  
    public void c1001()
    {
	MemoryMatrix mm;

	int m, i;
	
	int[] recipesToLaunch;
	recipesToLaunch = new int[10]; 

	// from pending specifications related to the recipe launching this computational step
	m = pendingComputationalSpecificationSet.getNumberOfMemoryMatrixesToBeUsed();
	if(m > eSFrameModelSwarm.getTotalMemoryMatrixNumber()
	   || m > 10)
	    {
		System.out.println("The number of types of the recipes launched by recipes\n"+
				   "cannot exceed 10 or the totalMemoryMatrixNumber.");
		MyExit.exit(1);
	    }

	try
	{    
	    BufferedReader fileRecipesFromRecipes = null;
	    StringTokenizer t;
	    String line = " "; 

	    fileRecipesFromRecipes = new BufferedReader(new 
				      FileReader("recipeData/recipesFromRecipes.txt"));

	    t = new StringTokenizer(line, " ");
	    for (i = 0; i < m; i++)
	    {	
		t=MyReader.check("recipeData/recipesFromRecipes.txt",
				 fileRecipesFromRecipes, line, t);
		recipesToLaunch[i] = Integer.parseInt(t.nextToken());
	    }

	    fileRecipesFromRecipes.close();
	}    
	catch (FileNotFoundException fnfe)
	{
	    System.out.println(fnfe);
	    MyExit.exit(1);
	} 
	catch (IOException e)
	{
	    System.out.println("IOException:" + e.toString());
	    MyExit.exit(1);
	}

	for (i = 0; i < m; i++)
	{    
	    mm =(MemoryMatrix) pendingComputationalSpecificationSet.
		getMemoryMatrixAddress(i);
	    layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	    mm.setValue(layer,0,0,recipesToLaunch[i]);
	}
	done=true;
    } // end c1001



    public void c1002()
    {
	MemoryMatrix mm;
	int aRecipe;


	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1)
	   {
	       System.out.println("Code -1002 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
	       MyExit.exit(1);
	   }

	mm =(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	
	aRecipe =(int) mm.getValue(layer,0,0);
	myOrderDistiller.setComputationalRecipe(aRecipe); // a recipe to be launched
	                                                  // at the end of this cycle

	done=true;
    } // end c1002



}



